<?php

namespace Nwb\S3;

use Aws\S3\S3Client;
use SebastianBergmann\Environment\Console;

class S3
{
    private $configS3;
    protected $s3RootFolder = '';
    const TAG_CONTENT_TYPE_KEY = 'ContentType';

    public function __construct(?ConfigS3 $configS3 = null)
    {
        $this->configS3 = ($configS3 instanceof ConfigS3) ? $configS3 : new ConfigS3();
    }

    public function getConfigS3()
    {
        return $this->configS3;
    }

    /**
     * Envoi un fichier vers un stockage d'objets S3.
     * Si l'objet existe déjà il sera écrasé sans avertissement.
     * @param string $filePath Le chemin du fichier à charger.
     * @param string $destinationPath Le chemin de destination du fichier sur Amazon S3.
     * @return \Aws\Result Le résultat de l'opération de chargement.
     */
    public function upload(string $filePath, string $destinationPath): \Aws\Result
    {
        $resource = fopen($filePath, 'r');
        if ($resource === false) {
            throw new \RuntimeException("Impossible d'ouvrir le fichier: $filePath");
        }

        try {
            return $this->uploadResource($resource, $destinationPath);
        } finally {
            fclose($resource);
        }
    }

    /**
     * Affiche le contenu d'un fichier S3 dans le navigateur.
     *
     * @param string $nomFichierS3 Le nom du fichier S3 à afficher.
     * @param bool $envoyerContentType Indique si l'en-tête 'Content-Type' doit être envoyé. Par défaut à true.
     * @return void
     */
    public function show(string $nomFichierS3, bool $envoyerContentType = true): void
    {
        $resource = $this->downloadResource($nomFichierS3);

        // Déterminez le Content-Type
        $contentType = $this->determineContentType($nomFichierS3, $resource);

        if ($envoyerContentType && $contentType) {
            header("Content-Type: $contentType");
        }

        // Affiche le contenu de la ressource
        fpassthru($resource);

        // Fermez la ressource
        fclose($resource);
    }

    /**
     * Télécharge un fichier depuis Amazon S3 et l'enregistre localement.
     *
     * @param string $nomFichierS3 Le nom du fichier à télécharger sur S3.
     * @param string $destinationPath Le chemin de destination où le fichier sera enregistré localement.
     * @return void
     */
    public function download(string $nomFichierS3, string $destinationPath): void
    {
        $nomFichierS3 = $this->normalizePath($nomFichierS3);
        $s3 = new S3Client($this->configS3->getConfig());
        $result = $s3->getObject([
            'Bucket' => $this->configS3->getBucket(),
            'Key' => $nomFichierS3,
        ]);
        file_put_contents($destinationPath, $result['Body']);
    }

    /**
     * Envoi une ressource vers un stockage d'objets S3.
     * Si l'objet existe déjà il sera écrasé sans avertissement.
     * @param resource $resource La ressource à charger.
     * @param string $destinationPath Le chemin de destination de la ressource sur Amazon S3.
     * @return \Aws\Result Le résultat de l'opération de chargement.
     */
    public function uploadResource($resource, string $destinationPath): \Aws\Result
    {
        if (!is_resource($resource)) {
            throw new \InvalidArgumentException("uploadResource() The provided resource is invalid.");
        }

        $data = stream_get_contents($resource);

        $tags = [];

        if ($this->configS3->getEncryptionKey()) {
            $data = $this->encryptData($data);

            // Si crypté, stockez le Content-Type dans les métadonnées ou les tags
            $contentType = mime_content_type($resource);

            $tags = [self::TAG_CONTENT_TYPE_KEY => $contentType];
        }

        $destinationPath = $this->normalizePath($destinationPath);
        $s3 = new S3Client($this->configS3->getConfig());
        $result = $s3->putObject([
            'Bucket' => $this->configS3->getBucket(),
            'Key' => $destinationPath,
            'Body' => $data
        ]);

        if (!empty($tags)) {
            $s3->putObjectTagging([
                'Bucket' => $this->configS3->getBucket(),
                'Key' => $destinationPath,
                'Tagging' => [
                    'TagSet' => array_map(function ($key, $value) {
                        return ['Key' => $key, 'Value' => $value];
                    }, array_keys($tags), $tags)
                ],
            ]);
        }

        return $result;
    }

    /**
     * Supprime un fichier du bucket S3.
     *
     * @param string $nomFichierS3 Le nom du fichier à supprimer.
     * @return \Aws\Result Le résultat de la suppression.
     */
    public function delete(string $nomFichierS3): \Aws\Result
    {
        $nomFichierS3 = $this->normalizePath($nomFichierS3);
        $s3 = new S3Client($this->configS3->getConfig());
        return $s3->deleteObject([
            'Bucket' => $this->configS3->getBucket(),
            'Key' => $nomFichierS3,
        ]);
    }

    /**
     * Normalisation du chemin d'accès S3
     * Supprime le slash au début du chemin si présent et ajoute le dossier racine (si défini)
     * @param string $chemin
     * @return string
     */
    private function normalizePath(string $chemin): string
    {
        if (substr($chemin, 0, 1) === '/') {
            $chemin = substr($chemin, 1);
        }
        return $this->s3RootFolder . $chemin;
    }

    /**
     * Renvoie vrai si le fichier existe sur S3
     * @param string $nomFichierS3
     * @return bool
     */
    public function fileExists(string $nomFichierS3): bool
    {
        $nomFichierS3 = $this->normalizePath($nomFichierS3);
        $s3 = new S3Client($this->configS3->getConfig());
        return $s3->doesObjectExist($this->configS3->getBucket(), $nomFichierS3);
    }

    /**
     * Liste les fichiers dans un chemin spécifié dans le bucket S3.
     * Peut lister récursivement si spécifié.
     *
     * @param string $path Le chemin à l'intérieur du bucket où lister les fichiers.
     * @param bool $recursive Indique si la liste doit être récursive.
     * @return array Un tableau contenant les noms des fichiers.
     */
    public function listFiles(string $path = '', bool $recursive = false): array
    {
        $path = $this->normalizePath($path);
        $s3 = new S3Client($this->configS3->getConfig());

        $fileList = [];
        $marker = null;

        do {
            $options = [
                'Bucket' => $this->configS3->getBucket(),
                'Prefix' => $path
            ];

            if (!$recursive) {
                $options['Delimiter'] = '/';
            }

            if ($marker) {
                $options['Marker'] = $marker;
            }

            $objects = $s3->listObjects($options);

            foreach ($objects['Contents'] as $object) {
                $fileList[] = $path . $object['Key'];
            }

            $marker = $objects['NextMarker'] ?? null;
        } while ($objects['IsTruncated']);

        return $fileList;
    }
    /**
     * Télécharge un fichier depuis Amazon S3 et retourne la ressource de flux.
     *
     * @param string $nomFichierS3 Le nom du fichier à télécharger sur S3.
     * @return resource La ressource de flux du fichier téléchargé.
     * @throws \Exception Si la récupération échoue.
     */
    public function downloadResource(string $nomFichierS3)
    {
        $nomFichierS3 = $this->normalizePath($nomFichierS3);
        $s3 = new S3Client($this->configS3->getConfig());

        try {
            $result = $s3->getObject([
                'Bucket' => $this->configS3->getBucket(),
                'Key' => $nomFichierS3,
            ]);
        } catch (\Aws\Exception\AwsException $e) {
            throw new \Exception("Erreur lors du téléchargement du fichier depuis S3: " . $e->getMessage());
        }

        $data = (string)$result['Body'];
        if ($this->configS3->getEncryptionKey()) {
            $data = $this->decryptData($data);
        }

        $stream = fopen('php://temp', 'r+');
        fwrite($stream, $data);
        rewind($stream);

        return $stream;
    }


    private function encryptData($data)
    {
        $key = $this->configS3->getEncryptionKey();
        if (!$key) {
            throw new \RuntimeException("Aucune clé de chiffrement définie.");
        }
        $iv = openssl_random_pseudo_bytes(openssl_cipher_iv_length('aes-256-cbc'));
        $encryptedData = openssl_encrypt($data, 'aes-256-cbc', $key, OPENSSL_RAW_DATA, $iv);

        // Concatène l'IV et les données chiffrées (sans Base64)
        return $iv . $encryptedData;
    }

    private function decryptData($data)
    {
        $key = $this->configS3->getEncryptionKey();
        if (!$key) {
            throw new \RuntimeException("Aucune clé de chiffrement définie.");
        }
        $ivSize = openssl_cipher_iv_length('aes-256-cbc');
        $iv = substr($data, 0, $ivSize);
        $encryptedData = substr($data, $ivSize);

        return openssl_decrypt($encryptedData, 'aes-256-cbc', $key, OPENSSL_RAW_DATA, $iv);
    }

    private function determineContentType(string $nomFichierS3, $resource)
    {
        if ($this->configS3->getEncryptionKey()) {
            $metadata = $this->getTags($nomFichierS3);
            return $metadata[self::TAG_CONTENT_TYPE_KEY] ?? null;
        }

        return mime_content_type($resource);
    }

    /**
     * Récupère les tags d'un fichier dans le bucket S3.
     *
     * @param string $nomFichierS3 Le nom du fichier dont on veut récupérer les tags.
     * @return array Les tags du fichier.
     * @throws \Exception Si la récupération échoue.
     */
    public function getTags(string $nomFichierS3): array
    {
        $nomFichierS3 = $this->normalizePath($nomFichierS3);
        $s3 = new S3Client($this->configS3->getConfig());

        try {
            // Récupération des tags
            $tagsResult = $s3->getObjectTagging([
                'Bucket' => $this->configS3->getBucket(),
                'Key' => $nomFichierS3,
            ]);
            $tags = $tagsResult->get('TagSet');

            // Transformer les tags en un tableau associatif plus simple
            $formattedTags = [];
            foreach ($tags as $tag) {
                $formattedTags[$tag['Key']] = $tag['Value'];
            }

            return $formattedTags;
        } catch (\Aws\Exception\AwsException $e) {
            throw new \Exception("Erreur lors de la récupération des tags du fichier depuis S3: " . $e->getMessage());
        }
    }
}
